/*
 *      pmftrans.c: PMF utility of primary transformation.
 *                 programmed by J. Terazono
 *   copyright (c) 1993 Jun-ya Terazono, all rights reserved.
 *
 */

/*
 *                         Outline of "pmftrans"
 *  This is a program for a primary transformation filtering of PMF format 
 *  image file. 
 *     Usage is:
 *    pmftrans [ parameters ] [ infile ] [ outfile ]
 *  
 *   If infile is omitted, stdin input is assumed.
 *   If outfile is omitted, stdout output is executed.
 */

/* ============================= Include files ============================= */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pmf.h>

/* =========================== Macro definitions =========================== */

#define  TRUE    1
#define  FALSE   0
#define  OutputError(x)  fprintf( stderr, (x))
#define  max(x,y)   ((x) > (y)) ? (x) : (y)
#define  min(x,y)   ((x) < (y)) ? (x) : (y)

/*
 *  Error number definitions 
 */

#define EPXNOERROR    0    /* program exited with no error */
#define EPXNOARGS     1    /* no arguments */
#define EPXILOPTS     2    /* illegal option character detected */
#define EPXILOPTFORM  3    /* illegal option format */
#define EPXZEROSTR    4    /* length of the string is zero */
#define EPXZERODET    5    /* determinant is zero */
#define EPXPMFLOAD    6    /* failure of PMFLoad */
#define EPXPMFCREATE  7    /* failure of PMFCreate */
#define EPXPMFSAVE    8    /* failure of PMFSave */
#define EPXERROR      9    /* other error */

static char *errormsg[] = {
    "Program exited with no error",                  /* EPXNOERROR */
    "No arguments",                                  /* EPXNOARGS */
    "Illegal option switch character detected",      /* EPXILOPTS */
    "Illegal option assignment",                     /* EPXILOPTFORM */
    "Null string detected",                          /* EPXZEROSTR */
    "Determinant is zero",                           /* EPXZERODET */
    "PMFLoad failed",                                /* EPXPMFLOAD */
    "PMFCreate failed",                              /* EPXPMFCREATE */
    "PMFSave failed",                                /* EPXPMFSAVE */
    "Other error ( reason unknown )"                 /* EPXERROR */
	};

/* ========================= Structure definitions ========================= */

typedef struct    {
    int  magnify;    /* magnification */
    int  rotate;     /* rotation */
    int  matrix;     /* direct matrix */
    int  infile;     /* input file */
    int  outfile;    /* output file */
    }   OPTS;

typedef struct   {
    double x11;
    double x12;
    double x21;
    double x22;
    }   MATRIX;

typedef struct    {
    short   width;
    short   height;
    } WIDHT;

typedef struct    {
    short   xmin;
    short   xmax;
    short   ymin;
    short   ymax;
    }  LIMIT;

typedef struct    {
    short   x;
    short   y;
    }  POINT;

/* ========================== Function definitions ========================= */
void CreateInvMatrix();
void CreateMagMatrix();
void CreateRotMatrix();
void DisplayError();
void DisplayUsage();
void SetLimit();
void SetMatrix();
void SetZeroPix();
void SpecifyArgs();

/* ========================== Program statements ========================== */
main( argc, argv )
int      argc;
char   **argv;
    {
    OPTS     opts;      /* options */
    MATRIX   matrix, inv;
    PMF     *pmfin, *pmfout;       /* PMF data structure */
    WIDHT    outwidht;
    LIMIT    limit;
    short      x, y;
    PMFPixel  pix, zeropix;
    POINT    org;
    double   temp1, temp2;

    opts.magnify = opts.infile = opts.outfile = FALSE;
    matrix.x11 = matrix.x22 = 1.0;  matrix.x12 = matrix.x21 = 0.0;

    SpecifyArgs( argc, argv, &opts, &matrix );

    if(( pmfin = PMFLoad( NULL )) == NULL )    {
	PMFperror( NULL );
	exit( EPXPMFLOAD );
	}

    SetZeroPix( pmfin->type, &zeropix );
    fprintf( stderr, "pmfin: type = %d, width = %d, height = %d\n",
	   pmfin->type, pmfin->width, pmfin->height );

    CreateInvMatrix( &matrix, &inv );
    SetLimit( &matrix, pmfin->width, pmfin->height, &limit );
    outwidht.width = limit.xmax - limit.xmin;
    outwidht.height = limit.ymax - limit.ymin;

    if( NULL == ( pmfout = PMFCreate( pmfin->type,
				     outwidht.width, outwidht.height )))    {
	PMFperror( NULL );
	exit( EPXPMFCREATE );
	}

    fprintf( stderr, "pmfout: type = %d, width = %d, height = %d\n",
	   pmfout->type, pmfout->width, pmfout->height );

    for( y = limit.ymin; y < limit.ymax; y++ )    {
	for( x = limit.xmin; x < limit.xmax; x++ )    {
	    temp1 = inv.x11 * (double)x + inv.x12 * (double)y;
	    temp2 = inv.x21 * (double)x + inv.x22 * (double)y;
	    org.x = (short)temp1;
	    org.y = (short)temp2;
	    if( org.x < 0 || org.x > pmfin->width ||
	       org.y < 0 || org.y > pmfin->height )
		pix = zeropix;
	    else 
		pix = PMFGetPixel( pmfin, org.x, pmfin->height - org.y );

	    PMFPutPixel( pmfout, x - limit.xmin,
			pmfout->height - ( y - limit.ymin ), pix );
	    }
	}

    if( PMFSave( NULL, pmfout, NULL ))    {
	PMFperror( NULL );
	exit( EPXPMFSAVE );
	}
    }

/* ===== Display error messages ===== */
void DisplayError( errno )
int   errno;
    {
    fprintf( stderr, "%s.\n", errormsg[errno] );
    DisplayUsage();
    exit( errno );
    }

/* ===== Display usage if necessary ===== */
void DisplayUsage()
    {
    OutputError( "usage: pmftrans [ parameter ] [ infile ] [ outfile ]\n" );
    }

/* ===== Specify arguments and sets matrix ===== */
void SpecifyArgs( argc, argv, option, matrix )
int         argc;
char      **argv;
OPTS       *option;
MATRIX     *matrix;    /* transformation matrix */
    {
    int    i;

    if( argc == 1 ) 
	DisplayError( EPXNOARGS );

    for( i = 1; i < argc; i++ )    {
	if( *argv[i] == '-' )
	    switch( *((argv[i])+1) )    {
	    case 'm':
		option->magnify = TRUE;
		CreateMagMatrix( matrix, argv[i+1] );	
		break;
	    case 'r':
		option->rotate = TRUE;
		CreateRotMatrix( matrix, argv[i+1] );
		break;
	    case 'x':
		option->matrix = TRUE;
		SetMatrix( matrix, argv[i+1] );
		break;
	    default:
		DisplayError( EPXILOPTS );
		}
	}
    }

/* ===== Create matrix for linear transformation ===== */
void CreateMagMatrix( matrix, specstr )
MATRIX    *matrix;         /* transformation matrix */
char      *specstr;        /* specification strings */
    {
    int   i,               /* loop counter */
          detect = FALSE,  /* detection notifier */
	  speclen;         /* length of specification strings */

    speclen = strlen( specstr );
    if( speclen <= 0 ) DisplayError( EPXZEROSTR );

    for( i = 0; i < speclen; i++ )    {
	if( specstr[i] == 'x' )    {
	    if( detect == FALSE )
		detect = TRUE;
	    else
		DisplayError( EPXILOPTFORM );
	    }
	}

    if( detect == TRUE )       /* [x]x[y] type assignment */
	sscanf( specstr, "%lfx%lf", &(matrix->x11), &(matrix->x22) );
    else                       /* [x] type assignment */
	matrix->x11 = matrix->x22 = atof( specstr );
    }

/* ===== Create rotation matrix ===== */
void CreateRotMatrix( matrix, specstr )
MATRIX    *matrix;         /* transformation matrix */
char      *specstr;        /* specification strings */
    {
    int     speclen;       /* length of specification strings */
    double  angledeg,      /* angle of rotation in degree */
            anglerad;      /* angle of rotation in radian */

    speclen = strlen( specstr );
    if( speclen <= 0 ) DisplayError( EPXZEROSTR );

    angledeg = atof( specstr );
    anglerad = angledeg / 180. * M_PI;

    matrix->x11 = matrix->x22 = cos( anglerad );
    matrix->x21 = sin( anglerad );
    matrix->x12 = -matrix->x21;
    }

/* ===== Calculate inversion matrix ===== */
void CreateInvMatrix( matrix, inv )
MATRIX    *matrix,  *inv;
    {
    double   det;       /* determinant */

    det = matrix->x11* matrix->x22 - matrix->x12 * matrix->x21;
    if( det == 0.0 ) exit( EPXZERODET );

    inv->x11 =  matrix->x22 / det;
    inv->x22 =  matrix->x11 / det;
    inv->x12 = -matrix->x12 / det;
    inv->x21 = -matrix->x21 / det;
    }

/* ===== Set limit of the coordinates ===== */
void SetLimit( matrix, width, height, limit )
MATRIX    *matrix;     /* transformation matrix */
short      width, height;
LIMIT     *limit;      /* limit of coordinates ( in cartesian ) */
    {
    POINT    rect[4]; /* rectangle points after transformation */
    int      i;
    double   temp1, temp2;

    limit->xmax = limit->ymax = -10000;
    limit->xmin = limit->ymin = 10000;

    temp1 = matrix->x11 * (double)width;
    temp2 = matrix->x21 * (double)width;
    rect[0].x = (short)temp1;
    rect[0].y = (short)temp2;

    temp1 = matrix->x12 * (double)height;
    temp2 = matrix->x22 * (double)height;
    rect[2].x = (short)temp1;
    rect[2].y = (short)temp2;

    rect[1].x = rect[0].x + rect[2].x;
    rect[1].y = rect[0].y + rect[2].y;

    rect[3].x = rect[3].y = 0;

    for( i = 0; i < 4; i++ )    {
	if( rect[i].x > limit->xmax ) limit->xmax = rect[i].x;
	if( rect[i].x < limit->xmin ) limit->xmin = rect[i].x;
	if( rect[i].y > limit->ymax ) limit->ymax = rect[i].y;
	if( rect[i].y < limit->ymin ) limit->ymin = rect[i].y;
	}
    }

/* ===== Set matrix from command line input ===== */
void SetMatrix( matrix, matstr )
MATRIX    *matrix;    /* set matrix */
char      *matstr;    /* command line argument */
    {
    int    i, matlen,
           counter = 0;

    matlen = strlen( matstr );
    if( matlen == 0 ) DisplayError( EPXZEROSTR );

    for( i = 0; i < matlen; i++ ) if( matstr[i] == ',' ) counter++;
    
    if( counter != 3 ) DisplayError( EPXILOPTFORM );

    sscanf( matstr, "%lf,%lf,%lf,%lf",
	   &(matrix->x11), &(matrix->x12), &(matrix->x21), &(matrix->x22) );
    }

/* ===== Set pixel structure "zeropix" ===== */
void SetZeroPix( type, zeropix )
short  type;
PMFPixel *zeropix;
    {
    switch( type )    {
    case PMF_BITMAP:
	zeropix->bit = 0;
	break;
    case PMF_GRAYMAP:
	zeropix->gray = 0;
	break;
    case PMF_PIXELMAP:
	zeropix->pixel.r = zeropix->pixel.g = zeropix->pixel.b =
	    (unsigned char)0;
	break;
    case PMF_ORIGINAL:
	break;
    default:
	exit( 2 );
	}
    }

/* =========================== END of the program ========================== */
